@@ -32,11 +32,19 @@ extern "C"
      *
      * @return The heuristic optimal matrix with bootstrapped standard deviations.
      */
-    Matrix aggregateGroups(const Matrix *xmat, const Matrix *wmat, int *results, int *cuts, bool *bestResult,
-                           double set_threshold, const char *set_method, bool feasible, int bootiter,
-                           const char *p_method, const char *q_method, const double convergence,
-                           const double log_convergence, const int maxIter, double maxSeconds, const bool verbose,
-                           QMethodInput *inputParams);
+    Matrix aggregateGroups(
+        // ---- Matrices
+        EMContext *ctx,
+
+        // ---- Results
+        int *results, // Array with cutting indices
+        int *cuts,    // Amount of cuts
+        bool *bestResult,
+
+        // ---- EM and Bootstrap parameters
+        double set_threshold, const char *set_method, bool feasible, int bootiter, const char *p_method,
+        const char *q_method, const double convergence, const double log_convergence, const int maxIter,
+        double maxSeconds, const bool verbose, QMethodInput *inputParams);
 
     /*
      * Function for testing all of the 2^{G-1} combinations, returning the best aggregation according the
@@ -65,8 +73,8 @@ extern "C"
                                      int bootiter, double max_qual, const char *p_method, const char *q_method,
                                      double convergence, double log_convergence, bool verbose, int maxIter,
                                      double maxSeconds, QMethodInput *inputParams, double *outBestLL, double **outBestQ,
-                                     Matrix **bestBootstrap, double *outBestTime, int *outFinishReason,
-                                     int *outIterTotal);
+                                     double **bestExpected, Matrix **bestBootstrap, double *outBestTime,
+                                     int *outFinishReason, int *outIterTotal);
 #ifdef __cplusplus
 }
 #endif
