@@ -44,8 +44,6 @@ SOFTWARE.
 #define Realloc(p, n, t) ((t *)R_chk_realloc((void *)(p), (size_t)((n) * sizeof(t))))
 #endif
 
-Set **HSETS = NULL;       // Global pointer to store all H sets
-Set **KSETS = NULL;       // Global pointer to store all K sets
 size_t **CANDIDATEARRAYS; // 2D array indexed by [c][b]
 
 /**
@@ -60,7 +58,6 @@ size_t **CANDIDATEARRAYS; // 2D array indexed by [c][b]
  * @return void
  *
  */
-
 void vectorDiff(const size_t *K, const size_t *H, size_t *arr)
 {
     for (uint16_t c = 0; c < TOTAL_CANDIDATES; c++)
@@ -133,9 +130,10 @@ bool ifAllElements(const size_t *hElement, const size_t *kElement)
  *
  */
 
-void generateConfigurations(int b, size_t *votes, int position, int remainingVotes, int numCandidates,
+void generateConfigurations(EMContext *ctx, int b, size_t *votes, int position, int remainingVotes, int numCandidates,
                             size_t ***results, size_t *count)
 {
+    Matrix *X = &ctx->X; // Get the X matrix
     // ---- Base case: we're on the last candidate ---- //
     if (position == numCandidates - 1)
     {
@@ -173,7 +171,7 @@ void generateConfigurations(int b, size_t *votes, int position, int remainingVot
             return;
         }
         // ---- Call the recursion ----
-        generateConfigurations(b, votes, position + 1, remainingVotes - i, numCandidates, results, count);
+        generateConfigurations(ctx, b, votes, position + 1, remainingVotes - i, numCandidates, results, count);
     }
     // ---...--- //
 }
@@ -194,7 +192,7 @@ void generateConfigurations(int b, size_t *votes, int position, int remainingVot
  * @return size_t **: A pointer that will store arrays of arrays, having all of the possible combinations.
  *
  */
-size_t **generateAllConfigurations(int b, int totalVotes, int numCandidates, size_t *count)
+size_t **generateAllConfigurations(EMContext *ctx, int b, int totalVotes, int numCandidates, size_t *count)
 {
     // ---- Initialize parameters ---- //
     size_t **results = NULL;
@@ -203,7 +201,7 @@ size_t **generateAllConfigurations(int b, int totalVotes, int numCandidates, siz
     // --- ... --- //
 
     // ---- Call the recursion ---- //
-    generateConfigurations(b, votes, 0, totalVotes, numCandidates, &results, count);
+    generateConfigurations(ctx, b, votes, 0, totalVotes, numCandidates, &results, count);
     // --- ... --- //
     Free(votes);
     return results;
@@ -218,34 +216,35 @@ size_t **generateAllConfigurations(int b, int totalVotes, int numCandidates, siz
  * @return void: Results written at the global variable HSETS.
  *
  */
-void generateHSets()
+void generateHSets(EMContext *ctx)
 {
-    // ---- Allocate memory for the `b` index ----
-    HSETS = Calloc(TOTAL_BALLOTS, Set *);
+    Matrix *W = &ctx->W; // Get the W matrix
+                         // ---- Allocate memory for the `b` index ----
+    ctx->hset = Calloc(TOTAL_BALLOTS * TOTAL_GROUPS, Set);
+    Set *HSETS = ctx->hset;
 
     for (uint32_t b = 0; b < TOTAL_BALLOTS; b++)
     { // ---- For every ballot box
 
         // ---- Allocate memory for the `g` index ----
-        HSETS[b] = Calloc(TOTAL_GROUPS, Set);
 
         for (uint16_t g = 0; g < TOTAL_GROUPS; g++)
         { // ---- For each group given a ballot box
 
             // ---- Initialize H set and set initial parameters ---- //
-            HSETS[b][g].b = b;
-            HSETS[b][g].g = g;
+            HSET(ctx, b, g)->b = b;
+            HSET(ctx, b, g)->g = g;
             // ---- Parameters for the function ----
             size_t total = (size_t)MATRIX_AT_PTR(W, b, g);
             // --- ... --- //
 
             // ---- Compute the set combinations ---- //
             size_t count = 0;
-            size_t **configurations = generateAllConfigurations(b, total, TOTAL_CANDIDATES, &count);
+            size_t **configurations = generateAllConfigurations(ctx, b, total, TOTAL_CANDIDATES, &count);
 
             // ---- Store configurations and size ----
-            HSETS[b][g].data = configurations;
-            HSETS[b][g].size = count;
+            HSET(ctx, b, g)->data = configurations;
+            HSET(ctx, b, g)->size = count;
             // --- ... --- //
         }
     }
@@ -261,23 +260,25 @@ void generateHSets()
  *
  */
 
-void generateKSets()
+void generateKSets(EMContext *ctx)
 {
+    Matrix *W = &ctx->W; // Get the W matrix
     // ---- Allocate memory for the `b` index ----
-    KSETS = Calloc(TOTAL_BALLOTS, Set *);
+    ctx->kset = Calloc(TOTAL_BALLOTS * TOTAL_GROUPS, Set);
+    Set *KSETS = ctx->kset;
 
     for (uint32_t b = 0; b < TOTAL_BALLOTS; b++)
     { // ---- For every ballot box
 
         // ---- Allocate memory for the `f` index ----
-        KSETS[b] = Calloc(TOTAL_GROUPS, Set);
+        // KSETS[b] = Calloc(TOTAL_GROUPS, Set);
 
         for (uint16_t g = 0; g < TOTAL_GROUPS; g++)
         { // ---- For each group given a ballot box
 
             // ---- Initialize K set and set initial parameters ---- //
-            KSETS[b][g].b = b;
-            KSETS[b][g].g = g;
+            KSET(ctx, b, g)->b = b;
+            KSET(ctx, b, g)->g = g;
             // ---- Parameters for the function ----
             size_t total = 0;
             for (uint16_t f = 0; f <= g; f++)
@@ -289,11 +290,11 @@ void generateKSets()
 
             // ---- Compute the set combinations ---- //
             size_t count = 0;
-            size_t **configurations = generateAllConfigurations(b, total, TOTAL_CANDIDATES, &count);
+            size_t **configurations = generateAllConfigurations(ctx, b, total, TOTAL_CANDIDATES, &count);
 
             // ---- Store configurations and size ----
-            KSETS[b][g].data = configurations;
-            KSETS[b][g].size = count;
+            KSET(ctx, b, g)->data = configurations;
+            KSET(ctx, b, g)->size = count;
             // --- ... --- //
         }
     }
@@ -359,8 +360,9 @@ double prod(const size_t *hElement, const Matrix *probabilities, const int f)
  * @return double: The result of the multinomial coefficient
  *
  */
-double multinomialCoeff(const int b, const int f, const size_t *hElement)
+double multinomialCoeff(EMContext *ctx, const int b, const int f, const size_t *hElement)
 {
+    Matrix *W = &ctx->W; // Get the W matrix
     // --- Compute ln(w_bf!). When adding by one, it considers the last element too ---
     double result = lgamma1p((int)MATRIX_AT_PTR(W, b, f));
 
@@ -387,9 +389,10 @@ double multinomialCoeff(const int b, const int f, const size_t *hElement)
  * @return The value of `A`
  *
  */
-double computeA(const int b, const int f, const size_t *hElement, const Matrix *probabilities)
+double computeA(EMContext *ctx, const int b, const int f, const size_t *hElement)
 {
-    return multinomialCoeff(b, f, hElement) * prod(hElement, probabilities, f);
+    Matrix *probabilities = &ctx->probabilities; // Get the probabilities matrix
+    return multinomialCoeff(ctx, b, f, hElement) * prod(hElement, probabilities, f);
 }
 
 /**
@@ -404,32 +407,37 @@ double computeA(const int b, const int f, const size_t *hElement, const Matrix *
  * @return void. Results to be written on the hash table
  *
  */
-void recursion(MemoizationTable *memo, const Matrix *probabilities)
+void recursion(EMContext *ctx, MemoizationTable *memo)
 {
-    // #pragma omp parallel for num_threads(3) schedule(static)
+    Matrix *probabilities = &ctx->probabilities; // Get the probabilities matrix
+    Matrix *W = &ctx->W;                         // Get the W matrix
+    Set *KSETS = ctx->kset;
+    Set *HSETS = ctx->hset;
+
     for (uint32_t b = 0; b < TOTAL_BALLOTS; b++)
-    {                   // ---- For each ballot box
-        if (b % 5 == 0) // Checks condition every 5 iterations
-            R_CheckUserInterrupt();
+    { // ---- For each ballot box
+        // if (b % 5 == 0) // Checks condition every 5 iterations
+        // R_CheckUserInterrupt();
         for (uint16_t f = 0; f < TOTAL_GROUPS; f++)
         { // ---- For each group, given a ballot box
             // #pragma omp parallel for collapse(2) It actually worsened the performance
-            for (size_t k = 0; k < KSETS[b][f].size; k++)
+            for (size_t k = 0; k < KSET(ctx, b, f)->size; k++)
             { // ---- For each element from the K_bf set
                 // ---- If there's no existing combination, skip the loop ----
-                if (!KSETS[b][f].data || !KSETS[b][f].data[k] || KSETS[b][f].data[k] == NULL)
+                if (!KSET(ctx, b, f)->data || !(KSET(ctx, b, f)->data[k]) || (KSET(ctx, b, f)->data[k]) == NULL)
                 {
                     continue;
                 }
 
                 // ---- Define the current element from the K set ----
-                size_t *currentK = KSETS[b][f].data[k];
+                // size_t *currentK = KSETS[b][f].data[k];
+                size_t *currentK = KSET(ctx, b, f)->data[k];
 
-                for (size_t h = 0; h < HSETS[b][f].size; h++)
+                for (size_t h = 0; h < HSET(ctx, b, f)->size; h++)
                 { // ---- For each element from the H_bf set
-                    if (HSETS[b][f].size > 5000 && h % 250 == 0)
-                        R_CheckUserInterrupt();
-                    size_t *currentH = HSETS[b][f].data[h];
+                    // if (HSET(ctx, b, f)->size > 5000 && h % 250 == 0)
+                    // R_CheckUserInterrupt();
+                    size_t *currentH = HSET(ctx, b, f)->data[h];
                     // ---- If the element from h isn't smaller than the one from k ----
                     // ---- Note that, when generating the H set, the restriction from candidate votes was also imposed,
                     // so it excluded "trivial" cases ----
@@ -440,7 +448,7 @@ void recursion(MemoizationTable *memo, const Matrix *probabilities)
 
                     // ---- Compute the values that are independent from c and g ---- //
                     // ---- The value `a` from the pseudocode. Check `computeA` for more information ----
-                    double a = computeA(b, f, currentH, probabilities);
+                    double a = computeA(ctx, b, f, currentH);
                     // ---- Initialize the variable that will store the past iteration ----
                     double valueBefore;
                     // ---- Substract the Kth element with the Hth element (k-h) ----
@@ -569,8 +577,12 @@ double exactLL(MemoizationTable *memo)
  * @note: A single pointer is used to store the array continously. This is for using cBLAS operations later.
  *
  */
-double *computeQExact(const Matrix *probabilities, QMethodInput params, double *ll)
+void computeQExact(EMContext *ctx, QMethodInput params, double *ll)
 {
+    Matrix *probabilities = &ctx->probabilities;
+    Matrix *X = &ctx->X;
+    Matrix *W = &ctx->W;
+    double *q = ctx->q;
 
     // ---- Initialize CANDIDATEARRAYS, which corresponds as a copy of `X` but in size_t. ---- //
     if (CANDIDATEARRAYS == NULL)
@@ -578,7 +590,6 @@ double *computeQExact(const Matrix *probabilities, QMethodInput params, double *
         // ---- Define the memory for the matrix and its parameters ----
         CANDIDATEARRAYS = Calloc(TOTAL_BALLOTS, size_t *);
         // ---- Parallelize the loop over the ballot boxes ----
-        // #pragma omp parallel for
         for (uint16_t b = 0; b < TOTAL_BALLOTS; b++)
         { // ---- For all ballot boxes
             // ---- Allocate memory for the candidate array ----
@@ -592,27 +603,13 @@ double *computeQExact(const Matrix *probabilities, QMethodInput params, double *
     }
     // --- ... --- //
 
-    // ---- Initialize the sets, to avoid unnecesary computations ---- //
-    if (HSETS == NULL && KSETS == NULL)
-    {
-        // ---- Note: both sets could had been created on the same loop at cost of code legibility. The reward is really
-        // small, arguably, doesn't make any difference. It was preffered to not sacrifice legibility ----
-        // ---- Create the H sets ----
-        generateHSets();
-        // ---- Create the K sets ----
-        generateKSets();
-    }
-    // --- ... --- //
-
     // ---- Initialize the dictionary variables ---- //
     // ---- Initialize the hash table ----
     MemoizationTable *table = initMemo();
-    // ---- Initialize the array to return
-    double *array2 = (double *)Calloc(TOTAL_BALLOTS * TOTAL_CANDIDATES * TOTAL_GROUPS, double);
     // --- ... --- //
 
     // ---- Start the main computation ---- //
-    recursion(table, probabilities);
+    recursion(ctx, table);
     // --- ... --- //
 
     // ---- Store the results ---- //
@@ -643,109 +640,19 @@ double *computeQExact(const Matrix *probabilities, QMethodInput params, double *
                 double result = num[c] / den;
                 if (!isnan(result) && !isinf(result))
                 {
-                    Q_3D(array2, b, g, c, (int)TOTAL_GROUPS, (int)TOTAL_CANDIDATES) = result;
+                    Q_3D(q, b, g, c, (int)TOTAL_GROUPS, (int)TOTAL_CANDIDATES) = result;
                 }
                 else
                 {
-                    Q_3D(array2, b, g, c, (int)TOTAL_GROUPS, (int)TOTAL_CANDIDATES) = 0;
+                    Q_3D(q, b, g, c, (int)TOTAL_GROUPS, (int)TOTAL_CANDIDATES) = 0;
                 }
 
             } // ---- End loop on candidates
         } // ---- End loop on groups
     } // ---- End loop on ballot boxes
 
-    *ll = exactLL(table);
+    if (params.computeLL)
+        *ll = exactLL(table);
     freeMemo(table);
-    return array2;
-    // ---...--- //
-}
-
-/**
- * @brief Frees al the memory allocated on the H set.
- *
- */
-void freeHSet(void)
-{
-    // ---- If the set was created ----
-    if (HSETS != NULL)
-    {
-        for (uint32_t b = 0; b < TOTAL_BALLOTS; b++)
-        { // ---- For each ballot box
-            if (HSETS[b] != NULL)
-            { // ---- If the value was created ----
-                for (uint16_t g = 0; g < TOTAL_GROUPS; g++)
-                { // ---- For each group given a ballot box
-                    if (HSETS[b][g].data != NULL)
-                    { // ---- If the array was created ----
-                        // ---- Free each configuration in the H set ----
-                        for (size_t i = 0; i < HSETS[b][g].size; i++)
-                        { // ---- For each combination in the bth and gth configuration ----
-                            Free(HSETS[b][g].data[i]);
-                        }
-                        Free(HSETS[b][g].data); // Free the array of configurations
-                    }
-                }
-                Free(HSETS[b]); // Free the array of Hsets for this ballot
-            }
-        }
-        Free(HSETS); // Free the global array
-        HSETS = NULL;
-    }
-}
-
-/**
- * @brief Frees al the memory allocated on the K set.
- *
- */
-void freeKSet(void)
-{
-    // ---- If the set was created ----
-    if (KSETS != NULL)
-    {
-        for (uint32_t b = 0; b < TOTAL_BALLOTS; b++)
-        { // ---- For each ballot box
-            if (KSETS[b] != NULL)
-            { // ---- If the value was created ----
-                for (uint16_t g = 0; g < TOTAL_GROUPS; g++)
-                { // ---- For each group given a ballot box
-                    if (KSETS[b][g].data != NULL)
-                    { // ---- If the array was created ----
-                        // ---- Free each configuration in the K set ----
-                        for (size_t i = 0; i < KSETS[b][g].size; i++)
-                        { // ---- For each combination in the bth and gth configuration ----
-                            Free(KSETS[b][g].data[i]);
-                        }
-                        Free(KSETS[b][g].data); // Free the array of configurations
-                    }
-                }
-                Free(KSETS[b]); // Free the array of K sets for this ballot
-            }
-        }
-        Free(KSETS); // Free the global array
-        KSETS = NULL;
-    }
-}
-
-// __attribute__((destructor))
-void cleanExact(void)
-{
-    // ---- Destroy the candidate array of size_t ---- //
-    // ---- If the array was created ----
-    if (CANDIDATEARRAYS != NULL)
-    {
-        for (uint32_t b = 0; b < TOTAL_BALLOTS; b++)
-        { // ---- For each ballot box
-            Free(CANDIDATEARRAYS[b]);
-        }
-        Free(CANDIDATEARRAYS);
-        CANDIDATEARRAYS = NULL;
-    }
-    // ---...--- //
-
-    // ---- Destroy every set ---- //
-    // ---- Destroy the array of precomputed H ----
-    freeHSet();
-    // ---- Destroy the array of precomputed H ----
-    freeKSet();
     // ---...--- //
 }
