@@ -20,6 +20,9 @@ OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 SOFTWARE.
 */
 
+#ifdef _OPENMP
+#include <omp.h>
+#endif
 #include "multivariate-cdf.h"
 #include <R.h>
 #include <R_ext/Memory.h>
@@ -28,9 +31,6 @@ SOFTWARE.
 #include <Rinternals.h>
 #include <Rmath.h>
 #include <math.h>
-#ifdef _OPENMP
-#include <omp.h>
-#endif
 #include <stdint.h>
 #include <stdlib.h>
 #include <string.h>
@@ -44,15 +44,138 @@ SOFTWARE.
 #define Free(p) R_chk_free((void *)(p))
 #endif
 
-/*
- * @brief Compute the Montecarlo approximation proposed by Alan genz towards the most recent method, using univariate
- * conditional with quasirandom numbers.
+// ---- Arena with reusable buffers ---- //
+typedef struct
+{
+    double *feature;    // size C
+    double *mu_row;     // size C-1
+    double *featureA;   // size C-1
+    double *featureB;   // size C-1
+    double *mc_results; // size C
+} ArenaCDF;
+
+/**
+ * @brief Create and allocate an ArenaCDF for dimension C.
+ */
+static ArenaCDF ArenaCDF_init(int C)
+{
+    ArenaCDF A = (ArenaCDF){0};
+    A.feature = (double *)Calloc(C, double);
+    A.mu_row = (double *)Calloc(C - 1, double);
+    A.featureA = (double *)Calloc(C - 1, double);
+    A.featureB = (double *)Calloc(C - 1, double);
+    A.mc_results = (double *)Calloc(C, double);
+    return A;
+}
+
+/**
+ * @brief Free the buffers of the ArenaCDF.
+ */
+static void ArenaCDF_free(ArenaCDF *A)
+{
+    if (!A)
+        return;
+    if (A->feature)
+        Free(A->feature);
+    if (A->mu_row)
+        Free(A->mu_row);
+    if (A->featureA)
+        Free(A->featureA);
+    if (A->featureB)
+        Free(A->featureB);
+    if (A->mc_results)
+        Free(A->mc_results);
+    memset(A, 0, sizeof(*A));
+}
+
+// ---- Helpers without dynamic allocation (avoid Calloc/Free in loops) ---- //
+
+/**
+ * @brief Copy column `col` of M into a preallocated buffer.
  *
- * Computes an heuristic from the first Multivariate CDF proposed. More details at the following paper:
+ * @param[in]  *M   Input matrix
+ * @param[in]  col  Column index
+ * @param[out] *out Buffer of size M->rows
+ */
+static inline void getColumn_into(const Matrix *M, int col, double *out)
+{
+    for (int r = 0; r < M->rows; r++)
+        out[r] = MATRIX_AT_PTR(M, r, col);
+}
+
+/**
+ * @brief Copy row `row` of M into a preallocated buffer.
+ *
+ * @param[in]  *M   Input matrix
+ * @param[in]  row  Row index
+ * @param[out] *out Buffer of size M->cols
+ */
+static inline void getRow_into(const Matrix *M, int row, double *out)
+{
+    for (int c = 0; c < M->cols; c++)
+        out[c] = MATRIX_AT_PTR(M, row, c);
+}
+
+// ---- PDF into the midpoint ---- //
+static double pdf_midpoint(const Matrix *cholesky, const double *mu, const double *lower, const double *upper, int d)
+{
+    // ---- Get the middle point ---- //
+    double mid[d];
+    for (int i = 0; i < d; i++)
+        mid[i] = 0.5 * (lower[i] + upper[i]);
+    // ---...--- //
+
+    // ---- diff = mid - mu ---- //
+    double diff[d];
+    for (int i = 0; i < d; i++)
+        diff[i] = mid[i] - mu[i];
+    // ---...--- //
+
+    // ---- L*z = diff ---- //
+    double z[d];
+    for (int i = 0; i < d; i++)
+    {
+        double lii = MATRIX_AT_PTR(cholesky, i, i);
+        if (!(lii > 0.0))
+            return 0.0; // mal condicionada
+        double s = diff[i];
+        for (int j = 0; j < i; j++)
+            s -= MATRIX_AT_PTR(cholesky, i, j) * z[j];
+        z[i] = s / lii;
+    }
+    // ---...--- //
+
+    // ---- ||z||^2 ---- //
+    double z2 = 0.0;
+    for (int i = 0; i < d; i++)
+        z2 += z[i] * z[i];
+    // ---...--- //
+
+    // ---- log(|Σ|^(1/2)) = sum(log(L_ii)) ---- //
+    double log_det_sqrt = 0.0;
+    for (int i = 0; i < d; i++)
+        log_det_sqrt += log(MATRIX_AT_PTR(cholesky, i, i));
+    // ---...--- //
+
+    // ---- φ(mid; μ,Σ) = (2π)^(-d/2) * exp(-0.5*||z||^2) / Π L_ii ---- //
+    double log_norm = -0.5 * d * log(M_2_PI) - log_det_sqrt;
+    double log_pdf = log_norm - 0.5 * z2;
+    double pdf = exp(log_pdf);
+
+    return isfinite(pdf) ? pdf : 0.0;
+}
+
+/*
+ * @brief Compute the Montecarlo approximation proposed by Alan genz towards the
+ * most recent method, using univariate conditional with quasirandom numbers.
+ *
+ * Computes an heuristic from the first Multivariate CDF proposed. More details
+ * at the following paper:
  * https://www.researchgate.net/publication/2463953_Numerical_Computation_Of_Multivariate_Normal_Probabilities
  *
- * Note that this method gives the option to impose an error threshold. So, it stops iterating if, either the maximum
- * iterations are accomplished or the error threshold is passed.
+ * Note that this method gives the option to impose an error threshold. So, it
+ * stops iterating if, either the maximum iterations are accomplished or the
+ * error threshold is passed.
  *
  * @param[in] *cholesky. The cholesky matrix of the current iteration.
  * @param[in] *lowerBounds. An array with the initial lower bounds.
@@ -95,7 +218,8 @@ double genzMontecarloNew(const Matrix *cholesky, const double *lowerBounds, cons
         // ---- Do the main loop ---- //
         for (int i = 1; i < mvnDim; i++)
         {
-            // ---- Note that the summatory is equivalent to $\sum_{j=1}^{i-1}c_{ij}*y_{j}$.
+            // ---- Note that the summatory is equivalent to
+            // $\sum_{j=1}^{i-1}c_{ij}*y_{j}$.
             summatory = 0;
             for (int j = 0; j < i; j++)
             {
@@ -125,13 +249,16 @@ double genzMontecarloNew(const Matrix *cholesky, const double *lowerBounds, cons
 /*
  * @brief Compute the Montecarlo approximation proposed by Alan genz
  *
- * Compute a `manual` Montecarlo simulation, proposed by Dr. Alan genz book "Computation of Multivariate Normal and t
- * Probabilities". All of the variable names and conventions are adopted towards his proposed algorithm aswell. Refer to
- * https://www.researchgate.net/publication/2463953_Numerical_Computation_Of_Multivariate_Normal_Probabilities for more
- * information.
+ * Compute a `manual` Montecarlo simulation, proposed by Dr. Alan genz book
+ * "Computation of Multivariate Normal and t Probabilities". All of the variable
+ * names and conventions are adopted towards his proposed algorithm aswell.
+ * Refer to
+ * https://www.researchgate.net/publication/2463953_Numerical_Computation_Of_Multivariate_Normal_Probabilities
+ * for more information.
  *
- * Note that this method gives the option to impose an error threshold. So, it stops iterating if, either the maximum
- * iterations are accomplished or the error threshold is passed.
+ * Note that this method gives the option to impose an error threshold. So, it
+ * stops iterating if, either the maximum iterations are accomplished or the
+ * error threshold is passed.
  *
  * @param[in] *cholesky. The cholesky matrix of the current iteration.
  * @param[in] *lowerBounds. An array with the initial lower bounds.
@@ -189,7 +316,8 @@ double genzMontecarlo(const Matrix *cholesky, const double *lowerBounds, const d
         {
             double draw = d[i - 1] + randomVector[i - 1] * (e[i - 1] - d[i - 1]);
             y[i - 1] = qnorm(draw, 0.0, 1.0, 1, 0);
-            // ---- Note that the summatory is equivalent to $\sum_{j=1}^{i-1}c_{ij}*y_{j}$.
+            // ---- Note that the summatory is equivalent to
+            // $\sum_{j=1}^{i-1}c_{ij}*y_{j}$.
 
             summatory = 0;
             for (int j = 0; j < i; j++)
@@ -226,17 +354,21 @@ double genzMontecarlo(const Matrix *cholesky, const double *lowerBounds, const d
 /*
  * @brief Calls the main function to start all of the Montecarlo process
  *
- * Calls the function with all of the parameters needed to get the Montecarlo simulation of the Multivariate CDF.
+ * Calls the function with all of the parameters needed to get the Montecarlo
+ * simulation of the Multivariate CDF.
  *
  * @param[in] *chol A matrix with the cholenksy values of the current group
  * @param[in] *mu An array with the average values of the current feature vector
- * @param[in] *lowerLimits An array with the lower bounds of the integral (defined by the hypercube)
- * @param[in] *upperLimits An array with the upper bounds of the integral (defined by the hypercube)
- * @param[in] mvnDim The dimensions of the multivariate normal. Usually it's C-1.
+ * @param[in] *lowerLimits An array with the lower bounds of the integral
+ * (defined by the hypercube)
+ * @param[in] *upperLimits An array with the upper bounds of the integral
+ * (defined by the hypercube)
+ * @param[in] mvnDim The dimensions of the multivariate normal. Usually it's
+ * C-1.
  * @param[in] maxSamples Amount of samples for the Montecarlo simulation.
  * @param[in] epsilon The error threshold used for the genz Montecarlo.
- * @param[in] *method The method for calculating the Montecarlo simulation. Currently available methods are `Plain`,
- * `Miser` and `Vegas`.
+ * @param[in] *method The method for calculating the Montecarlo simulation.
+ * Currently available methods are `Plain`, `Miser` and `Vegas`.
  *
  * @note Refer to https://www.gnu.org/software/gsl/doc/html/montecarlo.html
  *
@@ -252,7 +384,9 @@ double Montecarlo(Matrix *chol, double *mu, double *lowerLimits, double *upperLi
     // ---- Case where there are no values ---- //
     if (MATRIX_AT_PTR(chol, 0, 0) == 0 ||
         memcmp(lowerLimits, upperLimits, sizeof(double) * (TOTAL_CANDIDATES - 1)) == 0)
+    {
         return 0;
+    }
 
     // ---- Perform integration ---- //
     if (strcmp(method, "genz") == 0)
@@ -267,9 +401,11 @@ double Montecarlo(Matrix *chol, double *mu, double *lowerLimits, double *upperLi
     }
     else
     {
-        error("Multivariate CDF: An invalid method was handed to the Montecarlo simulation for calculating the "
+        error("Multivariate CDF: An invalid method was handed to the Montecarlo "
+              "simulation for calculating the "
               "Multivariate CDF "
-              "integral.\nThe method handed is:\t%s\nThe current available methods are `genz` or `genz2`"
+              "integral.\nThe method handed is:\t%s\nThe current available methods "
+              "are `genz` or `genz2`"
               ".\n",
               method);
     }
@@ -277,158 +413,171 @@ double Montecarlo(Matrix *chol, double *mu, double *lowerLimits, double *upperLi
 }
 
 /*
- * @brief Gets the a matrix of mu values and the inverse sigma for a given ballot box
+ * @brief Gets the a matrix of mu values and the inverse sigma for a given
+ * ballot box
  *
  * @param[in] b The ballot box index.
- * @param[in] *probabilitiesReduced Probabilities matrix without the last candidate.
- * @param[in, out] **cholesky An array of `g` matrices that stores (`c-1`X`c-1`) matrices with the cholesky values. The
- * cholesky matrix is filled on the upper and lower triangle with the same values since it's simmetric.
+ * @param[in] *probabilitiesReduced Probabilities matrix without the last
+ * candidate.
+ * @param[in, out] **cholesky An array of `g` matrices that stores (`c-1`X`c-1`)
+ * matrices with the cholesky values. The cholesky matrix is filled on the upper
+ * and lower triangle with the same values since it's simmetric.
  * @param[in, out] *mu A matrix of size (`g`x`c-1`) with the averages per group.
  *
  * @return void. Results to be written on cholesky and mu
  */
-void getMainParameters(int b, Matrix const probabilitiesReduced, Matrix **cholesky, Matrix *mu)
+static void getMainParameters(EMContext *ctx, int b, const Matrix probabilitiesReduced, Matrix **cholesky, Matrix *mu)
 {
+    // ---- Compute mu and sigma ---- //
+    getAverageConditional(ctx, b, &probabilitiesReduced, mu, cholesky);
 
-    // --- Initialize empty array --- //
-    for (uint16_t g = 0; g < TOTAL_GROUPS; g++)
-    { // ---- For each group ----
-        cholesky[g] = (Matrix *)Calloc(1, Matrix);
-        *cholesky[g] = createMatrix(TOTAL_CANDIDATES - 1, TOTAL_CANDIDATES - 1); // Initialize
+    // ---- Cholesky factorization (if applicable) ---- //
+    if (TOTAL_CANDIDATES != 2)
+    {
+        for (uint16_t g = 0; g < TOTAL_GROUPS; g++)
+            choleskyMat(cholesky[g]);
     }
-    // ---...--- //
-
-    // ---- Get mu and sigma ---- //
-    getAverageConditional(b, &probabilitiesReduced, mu, cholesky);
-    for (uint16_t g = 0; g < TOTAL_GROUPS && TOTAL_CANDIDATES != 2; g++)
-    { // ---- For each group ----
-        choleskyMat(cholesky[g]);
-    }
-
-    // ---...--- //
 }
 
 /*
  * @brief Computes the `q` values from the multivariate CDF method
  *
  * @param[in] *probabilities A pointer towards the current probabilities matrix.
- * @param[in] monteCarloSamples The amount of samples to use in the Monte Carlo simulation
+ * @param[in] monteCarloSamples The amount of samples to use in the Monte Carlo
+ * simulation
  * @param[in] epsilon The error threshold used for the genz Montecarlo method.
- * @param[in] *method The method for calculating the Montecarlo simulation. Currently available methods are `Plain`,
- * `Miser` and `Vegas`.
+ * @param[in] *method The method for calculating the Montecarlo simulation.
+ * Currently available methods are `Genz` and `Genz1`.
  *
  * @return A contiguos array with all the new probabilities
  */
-double *computeQMultivariateCDF(Matrix const *probabilities, QMethodInput params, double *ll)
+void computeQMultivariateCDF(EMContext *ctx, QMethodInput params, double *ll)
 {
     *ll = 0.0;
+
+    Matrix *X = &ctx->X;
+    double *q = ctx->q;
+    Matrix *P = &ctx->probabilities;
+
+    const int B = (int)ctx->B;
+    const int G = (int)ctx->G;
+    const int C = (int)ctx->C;
+    const int d = C - 1;
+
     int monteCarloSamples = params.monteCarloIter;
     double epsilon = params.errorThreshold;
     const char *method = params.simulationMethod;
 
-    // ---- Define initial variables ---- //
-    Matrix probabilitiesReduced = removeLastColumn(probabilities);
-    double *array2 = (double *)Calloc(TOTAL_BALLOTS * TOTAL_CANDIDATES * TOTAL_GROUPS, double); // Array to return
-    double *logArray = (double *)Calloc(TOTAL_BALLOTS, double);
-    // --- ... --- //
+    // ---- Remove last column (needed for getAverageConditional) ---- //
+    Matrix P_red = removeLastColumn(P);
 
-    for (uint32_t b = 0; b < TOTAL_BALLOTS; b++)
-    { // --- For each ballot box
-        // ---- Get the values of the Multivariate CDF that only depends on `b` ---- //
-        // ---- Mu and inverse Sigma matrix ----
-        Matrix mu = createMatrix(TOTAL_GROUPS, TOTAL_CANDIDATES - 1);
-        Matrix **choleskyVals = (Matrix **)Calloc(TOTAL_GROUPS, Matrix *);
-        getMainParameters(b, probabilitiesReduced, choleskyVals, &mu);
-        // ---- Array with the results of the Xth candidate on ballot B ----
-        double *feature = getColumn(X, b); // Of size C-1
-                                           // ---...--- //
+    // ---- Optional log-likelihood array ---- //
+    double *logArray = NULL;
+    if (params.computeLL)
+        logArray = (double *)Calloc(B, double);
 
-        for (uint16_t g = 0; g < TOTAL_GROUPS; g++)
-        { // --- For each group given a ballot box
-            // ---- Define the current values to use that only depends on `g` ---- //
-            Matrix *currentCholesky = choleskyVals[g];
+    // ---- Arena reused across all ballots ---- //
+    ArenaCDF A = ArenaCDF_init(C);
 
-            double *currentMu = getRow(&mu, g);
-            // ---- Initialize empty variables to be filled ----
-            double montecarloResults[TOTAL_CANDIDATES];
-            double denominator = 0;
-            // ---...--- //
+    for (uint32_t b = 0; b < (uint32_t)B; b++)
+    {
+        // ---- Create per-ballot structures ---- //
+        Matrix **L = (Matrix **)Calloc(G, Matrix *);
+        for (int g = 0; g < G; g++)
+        {
+            L[g] = (Matrix *)Calloc(1, Matrix);
+            *L[g] = createMatrix(d, d); // The cholesky matrix of dimension C-1 x C - 1
+        }
+        Matrix mu = createMatrix(G, d); // Mu matrix of dimension G x C-1
 
-            for (uint16_t c = 0; c < TOTAL_CANDIDATES; c++)
-            { // --- For each candidate given a group and a ballot box
+        // ---- Fill mu and cholesky (no alloc inside) ---- //
+        getMainParameters(ctx, (int)b, P_red, L, &mu);
 
-                // ---- Define the borders of the hypercube ---- //
-                // ---- First, make a copy of the feature vector ----
-                double *featureCopyA = (double *)Calloc((TOTAL_CANDIDATES - 1), double);
-                double *featureCopyB = (double *)Calloc((TOTAL_CANDIDATES - 1), double);
+        // ---- feature = X[, b] into Arena buffer ---- //
+        getColumn_into(X, (int)b, A.feature);
 
-                memcpy(featureCopyA, feature, (TOTAL_CANDIDATES - 1) * sizeof(double));
-                memcpy(featureCopyB, feature, (TOTAL_CANDIDATES - 1) * sizeof(double));
+        for (uint16_t g = 0; g < G; g++)
+        {
+            Matrix *Lg = L[g];
+            getRow_into(&mu, g, A.mu_row);
 
-                // ---- Substract/add and normalize ----
-                // ---- Note that the bounds NEEDS to be standarized ----
-                // ---- $$a=\sigma^{-1}(a-\mu)$$ ----
-                for (uint16_t k = 0; k < TOTAL_CANDIDATES - 1; k++)
-                { // --- For each candidate coordinate that is going to be integrated
-                    featureCopyA[k] -= 0.5;
-                    featureCopyB[k] += 0.5;
+            double denom = 0.0;
+
+            for (uint16_t c = 0; c < C; c++)
+            {
+                // ---- Copy and adjust hypercube bounds ---- //
+                memcpy(A.featureA, A.feature, d * sizeof(double));
+                memcpy(A.featureB, A.feature, d * sizeof(double));
+                for (uint16_t k = 0; k < d; k++)
+                {
+                    A.featureA[k] -= 0.5;
+                    A.featureB[k] += 0.5;
                     if (k == c)
                     {
-                        featureCopyA[k] -= 1.0;
-                        featureCopyB[k] -= 1.0;
+                        A.featureA[k] -= 1.0;
+                        A.featureB[k] -= 1.0;
+                    }
+                    if (C == 2)
+                    {
+                        A.featureA[k] -= A.mu_row[k];
+                        A.featureB[k] -= A.mu_row[k];
                     }
-                    featureCopyA[k] -= currentMu[k];
-                    featureCopyB[k] -= currentMu[k];
                 }
-                // ---...--- //
-                // ---- Save the results and add them to the denominator ---- //
-                if (TOTAL_CANDIDATES != 2)
-                    montecarloResults[c] = Montecarlo(currentCholesky, currentMu, featureCopyA, featureCopyB,
-                                                      (int)TOTAL_CANDIDATES - 1, monteCarloSamples, epsilon, method);
 
+                // ---- Monte Carlo integral or closed form ---- //
+                double val;
+                if (C != 2)
+                {
+                    val = Montecarlo(Lg, A.mu_row, A.featureA, A.featureB, d, monteCarloSamples, epsilon, method);
+                }
                 else
                 {
-                    montecarloResults[c] =
-                        (pnorm(featureCopyB[0], 0.0, sqrt(MATRIX_AT_PTR(currentCholesky, 0, 0)), 1, 0) -
-                         pnorm(featureCopyA[0], 0.0, sqrt(MATRIX_AT_PTR(currentCholesky, 0, 0)), 1, 0));
+                    double s = sqrt(MATRIX_AT_PTR(Lg, 0, 0));
+                    val = pnorm(A.featureB[0], 0.0, s, 1, 0) - pnorm(A.featureA[0], 0.0, s, 1, 0);
                 }
-                montecarloResults[c] *= MATRIX_AT_PTR(probabilities, g, c);
-                denominator += !isnan(montecarloResults[c]) ? montecarloResults[c] : 0;
-                logArray[b] += g == 0 && !isnan(montecarloResults[c]) ? montecarloResults[c] : 0;
-                // TODO: Make an arena for this loop
-                Free(featureCopyA);
-                Free(featureCopyB);
-                // ---...--- //
-            } // --- End c loop
-            Free(currentMu);
-            freeMatrix(currentCholesky);
-            freeMatrix(choleskyVals[g]);
-            Free(choleskyVals[g]);
 
-            // ---- Add the final results to the array ----//
-            for (uint16_t c = 0; c < TOTAL_CANDIDATES; c++)
-            { // --- For each candidate
-                double result = montecarloResults[c] / denominator;
-                Q_3D(array2, b, g, c, (int)TOTAL_GROUPS, (int)TOTAL_CANDIDATES) =
-                    !isnan(result) && !isinf(result) ? result : 0;
-            } // --- End c loop
-            // ---...--- //
-        } // --- End g loop
-        // error("stopping here"); // Erse this
-        Free(feature);
+                // ---- Fallback with PDF midpoint ---- //
+                if (!(val > 0.0) || !isfinite(val))
+                    val = pdf_midpoint(Lg, A.mu_row, A.featureA, A.featureB, d);
+
+                // ---- Weight by prior probability ---- //
+                val *= MATRIX_AT_PTR(P, g, c);
+
+                A.mc_results[c] = val;
+                denom += isfinite(val) ? val : 0.0;
+
+                if (params.computeLL && g == 0 && isfinite(val))
+                    logArray[b] += val;
+            }
+
+            // ---- Normalize and store q ---- //
+            for (uint16_t c = 0; c < C; c++)
+            {
+                double result = A.mc_results[c] / denom;
+                Q_3D(q, b, g, c, G, C) = (isfinite(result) ? result : 0.0);
+            }
+        }
+
+        // ---- Free per-ballot matrices ---- //
+        for (int g = 0; g < G; g++)
+        {
+            freeMatrix(L[g]);
+            Free(L[g]);
+        }
+        Free(L);
         freeMatrix(&mu);
-        Free(choleskyVals);
-    } // --- End b loop
-    freeMatrix(&probabilitiesReduced);
-
-    // Compute the log-likelihood
-    double finalLikelihood = 0;
-    for (uint32_t b = 0; b < TOTAL_BALLOTS; b++)
-    {
-        finalLikelihood += logArray[b] != 0 ? log(logArray[b]) : 0;
     }
-    *ll = finalLikelihood;
-    // *ll = log(*ll);
-    free(logArray);
-    return array2;
+
+    // ---- Cleanup global ---- //
+    freeMatrix(&P_red);
+    ArenaCDF_free(&A);
+
+    if (params.computeLL)
+    {
+        double finalLL = 0.0;
+        for (uint32_t b = 0; b < (uint32_t)B; b++)
+            finalLL += (logArray[b] != 0.0) ? log(logArray[b]) : 0.0;
+        *ll = finalLL;
+        Free(logArray);
+    }
 }
